/*
 * Clock and power management. This exposes the combined functionality
 * provided by the X1600 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-x1600.h"
#include <math.h>



enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR
  Ddr_divider         = 0x02c,  // DDRCDR
  Mac_divider         = 0x054,  // MACCDR
  I2s_divider0        = 0x060,  // I2SCDR
  I2s_divider1        = 0x070,  // I2S1CDR
  Lcd_divider         = 0x064,  // LPCDR
  Msc_divider0        = 0x068,  // MSC0CDR
  Msc_divider1        = 0x0a4,  // MSC1CDR
  Sfc_divider         = 0x074,  // SFCCDR
  Ssi_divider         = 0x05c,  // SSICDR
  Cim_divider         = 0x078,  // CIMCDR
  Pwm_divider         = 0x06c,  // PWMCDR
  Can_divider0        = 0x0a0,  // CAN0CDR
  Can_divider1        = 0x0a8,  // CAN1CDR
  Cdbus_divider       = 0x0ac,  // CDBUSCDR
  Macphy0_divider     = 0x0e4,  // MPHY0C
  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_swi             = 0x0bc,  // CPM_SFTINT
  Ddr_gate            = 0x0d0,  // DRCG
  Cpm_scratch_prot    = 0x038,  // CPSPPR
  Cpm_scratch         = 0x034,  // CPSPR
  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1
  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_fraction_A      = 0x084,  // CPAPACR
  Pll_fraction_M      = 0x088,  // CPMPACR
  Pll_fraction_E      = 0x08c,  // CPEPACR
};

enum Clock_bits : unsigned
{
  // Clock_control

  Clock_gate_A              = 23, // GATE_SCLKA
  Clock_cpu_change_enable   = 22, // CE_CPU
  Clock_ahb0_change_enable  = 21, // CE_AHB0
  Clock_ahb2_change_enable  = 20, // CE_AHB2
  Clock_pclock_divider      = 16, // PDIV (slow APB peripherals)
  Clock_hclock2_divider     = 12, // H2DIV (fast AHB peripherals)
  Clock_hclock0_divider     = 8,  // H0DIV (fast AHB peripherals)
  Clock_l2cache_divider     = 4,  // L2CDIV
  Clock_cpu_divider         = 0,  // CDIV
};

enum Clock_source_bits : unsigned
{
  // Clock_control

  Clock_source_main     = 30, // SEL_SRC (output to SCLK_A)
  Clock_source_cpu      = 28, // SEL_CPLL (output to CCLK)
  Clock_source_hclock0  = 26, // SEL_H0PLL (output to AHB0)
  Clock_source_hclock2  = 24, // SEL_H2PLL (output to AHB2)

  // Ddr_divider

  Clock_source_ddr      = 30, // DCS

  // I2s_divider0

  Clock_source_i2s      = 31, // I2PCS

  // Lcd_divider

  Clock_source_lcd      = 30, // LPCS

  // Mac_divider

  Clock_source_mac      = 30, // MACPCS

  // Msc_divider0, Msc_divider1

  Clock_source_msc0     = 30, // MPCS
  Clock_source_msc1     = 30, // MPCS

  // Sfc_divider

  Clock_source_sfc      = 30, // SFCS

  // Ssi_divider

  Clock_source_ssi      = 30, // SPCS

  // Cim_divider

  Clock_source_cim      = 30, // CIMPCS

  // Pwm_divider

  Clock_source_pwm      = 30, // PWMPCS

  // Can_divider0, Can_divider1

  Clock_source_can0     = 30, // CA0CS
  Clock_source_can1     = 30, // CA1CS

  // Cdbus_divider

  Clock_source_cdbus    = 30, // CDCS
};

enum Clock_sources : unsigned
{
  // Clock_source_main

  Source_external       = 1,  // EXCLK
  Source_pll_A          = 2,  // APLL

  // Stoppable clock sources:
  // Clock_source_cpu, Clock_source_hclock0, Clock_source_hclock2,
  // Clock_source_ddr

  Source_mux_stopped    = 0,
  Source_mux_main       = 1,  // SCLK_A
  Source_mux_pll_M      = 2,  // MPLL

  // Unstoppable clock sources:
  // Clock_source_mac, Clock_source_i2s, Clock_source_lcd, Clock_source_msc0,
  // Clock_source_msc1, Clock_source_sfc, Clock_source_ssi, Clock_source_cim,
  // Clock_source_pwm, Clock_source_can0, Clock_source_can1, Clock_source_cdbus

  Source_main           = 0,  // SCLK_A
  Source_pll_M          = 1,  // MPLL
  Source_pll_E          = 2,  // EPLL

  Source_i2s_pll_E      = 1,  // EPLL

  Source_can_external   = 3,  // EXCLK
};

enum Clock_gate_bits : unsigned
{
  // Clock_gate0

  Clock_gate_ddr        = 31, // DDR
  Clock_gate_ahb0       = 29, // AHB0
  Clock_gate_apb0       = 28, // APB0
  Clock_gate_rtc        = 27, // RTC
  Clock_gate_aes        = 24, // AES
  Clock_gate_lcd_pixel  = 23, // LCD
  Clock_gate_cim        = 22, // CIM
  Clock_gate_dma        = 21, // PDMA
  Clock_gate_ost        = 20, // OST
  Clock_gate_ssi0       = 19, // SSI0
  Clock_gate_timer      = 18, // TCU
  Clock_gate_dtrng      = 17, // DTRNG
  Clock_gate_uart2      = 16, // UART2
  Clock_gate_uart1      = 15, // UART1
  Clock_gate_uart0      = 14, // UART0
  Clock_gate_sadc       = 13, // SADC
  Clock_gate_audio      = 11, // AUDIO
  Clock_gate_ssi_slv    = 10, // SSI_SLV
  Clock_gate_i2c1       = 8,  // I2C1
  Clock_gate_i2c0       = 7,  // I2C0
  Clock_gate_msc1       = 5,  // MSC1
  Clock_gate_msc0       = 4,  // MSC0
  Clock_gate_otg        = 3,  // OTG
  Clock_gate_sfc        = 2,  // SFC
  Clock_gate_efuse      = 1,  // EFUSE
  Clock_gate_nemc       = 0,  // NEMC

  // Clock_gate1

  Clock_gate_arb        = 30, // ARB
  Clock_gate_mipi_csi   = 28, // MIPI_CSI
  Clock_gate_intc       = 26, // INTC
  Clock_gate_gmac0      = 23, // GMAC0
  Clock_gate_uart3      = 16, // UART3
  Clock_gate_i2s0_tx    = 9,  // I2S0_dev_tclk
  Clock_gate_i2s0_rx    = 8,  // I2S0_dev_rclk
  Clock_gate_hash       = 6,  // HASH
  Clock_gate_pwm        = 5,  // PWM
  Clock_gate_cdbus      = 2,  // CDBUS
  Clock_gate_can1       = 1,  // CAN1
  Clock_gate_can0       = 0,  // CAN0

  // Special value

  Clock_gate_undefined  = 32,
};

// Clock gate register correspondences.

static uint32_t clock_gate_reg[Clock_identifier_count] = {
  /* Clock_aic_bitclk */  0,
  /* Clock_aic_pclk */    0,
  /* Clock_can0 */        Clock_gate1,
  /* Clock_can1 */        Clock_gate1,
  /* Clock_cdbus */       Clock_gate1,
  /* Clock_cim */         Clock_gate0,
  /* Clock_ddr */         Clock_gate0,
  /* Clock_dma */         Clock_gate0,
  /* Clock_emac */        0,
  /* Clock_hdmi */        0,
  /* Clock_i2c */         Clock_gate0,
  /* Clock_i2c0 */        Clock_gate0,
  /* Clock_i2c1 */        Clock_gate0,
  /* Clock_i2s */         0,
  /* Clock_i2s0_rx */     Clock_gate1,
  /* Clock_i2s0_tx */     Clock_gate1,
  /* Clock_kbc */         0,
  /* Clock_lcd */         0,
  /* Clock_lcd_pixel */   Clock_gate0,
  /* Clock_mac */         Clock_gate1,
  /* Clock_msc */         Clock_gate0,
  /* Clock_msc0 */        Clock_gate0,
  /* Clock_msc1 */        Clock_gate0,
  /* Clock_pwm */         Clock_gate1,
  /* Clock_pwm0 */        Clock_gate1,
  /* Clock_pwm1 */        0,
  /* Clock_scc */         0,
  /* Clock_sfc */         Clock_gate0,
  /* Clock_smb0 */        0,
  /* Clock_smb1 */        0,
  /* Clock_smb2 */        0,
  /* Clock_smb3 */        0,
  /* Clock_smb4 */        0,
  /* Clock_ssi */         Clock_gate0,
  /* Clock_timer */       Clock_gate0,
  /* Clock_uart0 */       Clock_gate0,
  /* Clock_uart1 */       Clock_gate0,
  /* Clock_uart2 */       Clock_gate0,
  /* Clock_uart3 */       Clock_gate1,
  /* Clock_udc */         0,
  /* Clock_uhc */         0,
  /* Clock_uprt */        0,
};

// Clock gate register bit correspondences.

static enum Clock_gate_bits clock_gate_bit[Clock_identifier_count] = {
  /* Clock_aic_bitclk */  Clock_gate_undefined,
  /* Clock_aic_pclk */    Clock_gate_undefined,
  /* Clock_can0 */        Clock_gate_can0,
  /* Clock_can1 */        Clock_gate_can1,
  /* Clock_cdbus */       Clock_gate_cdbus,
  /* Clock_cim */         Clock_gate_cim,
  /* Clock_ddr */         Clock_gate_ddr,
  /* Clock_dma */         Clock_gate_dma,
  /* Clock_emac */        Clock_gate_undefined,
  /* Clock_hdmi */        Clock_gate_undefined,
  /* Clock_i2c */         Clock_gate_i2c0,
  /* Clock_i2c0 */        Clock_gate_i2c0,
  /* Clock_i2c1 */        Clock_gate_i2c1,
  /* Clock_i2s */         Clock_gate_undefined,
  /* Clock_i2s0_rx */     Clock_gate_i2s0_rx,
  /* Clock_i2s0_tx */     Clock_gate_i2s0_tx,
  /* Clock_kbc */         Clock_gate_undefined,
  /* Clock_lcd */         Clock_gate_undefined,
  /* Clock_lcd_pixel */   Clock_gate_lcd_pixel,
  /* Clock_mac */         Clock_gate_gmac0,
  /* Clock_msc */         Clock_gate_msc0,
  /* Clock_msc0 */        Clock_gate_msc0,
  /* Clock_msc1 */        Clock_gate_msc1,
  /* Clock_pwm */         Clock_gate_pwm,
  /* Clock_pwm0 */        Clock_gate_pwm,
  /* Clock_pwm1 */        Clock_gate_undefined,
  /* Clock_scc */         Clock_gate_undefined,
  /* Clock_sfc */         Clock_gate_sfc,
  /* Clock_smb0 */        Clock_gate_undefined,
  /* Clock_smb1 */        Clock_gate_undefined,
  /* Clock_smb2 */        Clock_gate_undefined,
  /* Clock_smb3 */        Clock_gate_undefined,
  /* Clock_smb4 */        Clock_gate_undefined,
  /* Clock_ssi */         Clock_gate_ssi0,
  /* Clock_timer */       Clock_gate_timer,
  /* Clock_uart0 */       Clock_gate_uart0,
  /* Clock_uart1 */       Clock_gate_uart1,
  /* Clock_uart2 */       Clock_gate_uart2,
  /* Clock_uart3 */       Clock_gate_uart3,
  /* Clock_udc */         Clock_gate_undefined,
  /* Clock_uhc */         Clock_gate_undefined,
  /* Clock_uprt */        Clock_gate_undefined,
};

enum Divider_bits : unsigned
{
  Ddr_divider_value     = 0,  // DDRCDR
  Lcd_divider_value     = 0,  // LPCDR
};

enum Clock_status_values : unsigned
{
  Lcd_change_enable     = 0x20000000, // CE_LCD
  Lcd_change_busy       = 0x10000000, // LCD_BUSY
  Lcd_clock_stop        = 0x08000000, // LCD_STOP
};

enum Pll_bits : unsigned
{
  // Pll_control_A, Pll_control_M, Pll_control_E

  Pll_multiplier        = 20, // xPLLM
  Pll_input_division    = 14, // xPLLN
  Pll_output_division1  = 11, // xPLLOD1
  Pll_output_division0  = 8,  // xPLLOD0
  Pll_stable            = 3,  // xPLL_ON
  Pll_enabled           = 0,  // xPLLEN
};

enum Pll_bypass_bits : unsigned
{
  Pll_bypass_A          = 30, // APLL_BP
  Pll_bypass_M          = 28, // MPLL_BP
  Pll_bypass_E          = 26, // EPLL_BP
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_x1600_chip::Cpm_x1600_chip(l4_addr_t addr, uint32_t exclk_freq)
: _exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // add_cid("cpm");
  // add_cid("cpm-x1600");
  // register_property("exclk_freq", &_exclk_freq);
}

// Clock/timer control.

uint32_t
Cpm_x1600_chip::get_clock_gate_register(enum Clock_identifiers clock)
{
  return clock_gate_reg[clock];
}

uint32_t
Cpm_x1600_chip::get_clock_gate_value(enum Clock_identifiers clock)
{
  return 1 << clock_gate_bit[clock];
}

int
Cpm_x1600_chip::have_clock(enum Clock_identifiers clock)
{
  return !(_regs[get_clock_gate_register(clock)] & get_clock_gate_value(clock));
}

void
Cpm_x1600_chip::start_clock(enum Clock_identifiers clock)
{
  uint32_t gate = get_clock_gate_register(clock);

  _regs[gate] = _regs[gate] & ~get_clock_gate_value(clock);
}

void
Cpm_x1600_chip::stop_clock(enum Clock_identifiers clock)
{
  uint32_t gate = get_clock_gate_register(clock);

  _regs[gate] = _regs[gate] | get_clock_gate_value(clock);
}



// Utility methods.

uint32_t
Cpm_x1600_chip::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_x1600_chip::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

// General clock divider access.

uint8_t
Cpm_x1600_chip::_get_divider(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return get_field(reg, mask, shift) + 1;
}



// PLL control.

// Return whether the PLL is stable.

int
Cpm_x1600_chip::have_pll(uint32_t pll_reg)
{
  return _regs[pll_reg] & (1 << Pll_stable);
}

int
Cpm_x1600_chip::pll_enabled(uint32_t pll_reg)
{
  return _regs[pll_reg] & (1 << Pll_enabled);
}

int
Cpm_x1600_chip::pll_bypassed(uint32_t pll_reg)
{
  uint32_t mask;

  switch (pll_reg)
  {
    case Pll_control_A: mask = (1 << Pll_bypass_A); break;
    case Pll_control_M: mask = (1 << Pll_bypass_M); break;
    case Pll_control_E: mask = (1 << Pll_bypass_E); break;
    default: mask = 0; break;
  }

  return _regs[Pll_control] & mask;
}

void
Cpm_x1600_chip::pll_enable(uint32_t pll_reg)
{
  _regs[pll_reg] = _regs[pll_reg] | (1 << Pll_enabled);
  while (!(_regs[pll_reg] & (1 << Pll_stable)));
}

void
Cpm_x1600_chip::pll_disable(uint32_t pll_reg)
{
  _regs[pll_reg] = _regs[pll_reg] & ~(1 << Pll_enabled);
  while (_regs[pll_reg] & (1 << Pll_stable));
}

// Feedback (13-bit) multiplier.

uint16_t
Cpm_x1600_chip::get_multiplier(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x1fff, Pll_multiplier) + 1;
}

void
Cpm_x1600_chip::set_multiplier(uint32_t pll_reg, uint16_t multiplier)
{
  set_field(pll_reg, 0x1fff, Pll_multiplier, multiplier - 1);
}

// Input (6-bit) divider.

uint8_t
Cpm_x1600_chip::get_input_division(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x3f, Pll_input_division) + 1;
}

void
Cpm_x1600_chip::set_input_division(uint32_t pll_reg, uint8_t divider)
{
  set_field(pll_reg, 0x3f, Pll_input_division, divider - 1);
}

// Output (dual 3-bit) dividers.

uint8_t
Cpm_x1600_chip::get_output_division(uint32_t pll_reg)
{
  uint8_t d0 = get_field(pll_reg, 0x07, Pll_output_division0);
  uint8_t d1 = get_field(pll_reg, 0x07, Pll_output_division1);

  return d0 * d1;
}

void
Cpm_x1600_chip::set_output_division(uint32_t pll_reg, uint8_t divider)
{
  // Assert 1 as a minimum.
  // Divider 0 must be less than or equal to divider 1.

  uint8_t d0 = (uint8_t) floor(sqrt(divider ? divider : 1));
  uint8_t d1 = divider / d0;

  set_field(pll_reg, 0x07, Pll_output_division0, d0);
  set_field(pll_reg, 0x07, Pll_output_division1, d1);
}

uint32_t
Cpm_x1600_chip::get_pll_frequency(uint32_t pll_reg)
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    return (_exclk_freq * get_multiplier(pll_reg)) /
           (get_input_division(pll_reg) * get_output_division(pll_reg));
  else
    return _exclk_freq;
}

void
Cpm_x1600_chip::set_pll_parameters(uint32_t pll_reg, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  set_multiplier(pll_reg, multiplier);
  set_input_division(pll_reg, in_divider);
  set_output_division(pll_reg, out_divider);

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    while (!have_pll(pll_reg));
}



// CPU clock (CCLK) divider.

uint8_t
Cpm_x1600_chip::get_cpu_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_cpu_divider);
}

// Fast peripheral clock (H0CLK) divider.

uint8_t
Cpm_x1600_chip::get_hclock0_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_hclock0_divider);
}

// Fast peripheral clock (H2CLK) divider.

uint8_t
Cpm_x1600_chip::get_hclock2_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_hclock2_divider);
}

// Slow peripheral clock (PCLK) divider.

uint8_t
Cpm_x1600_chip::get_pclock_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_pclock_divider);
}

// LCD clock (LPCLK) divider for LCD pixel clock.

uint8_t
Cpm_x1600_chip::get_lcd_pixel_divider(uint8_t controller)
{
  (void) controller;
  return _get_divider(Lcd_divider, 0xff, Lcd_divider_value);
}

// Memory clock (DDR_CLK) divider.

uint8_t
Cpm_x1600_chip::get_memory_divider()
{
  return _get_divider(Ddr_divider, 0xf, Ddr_divider_value);
}

// LCD pixel clock divider.

void
Cpm_x1600_chip::set_lcd_pixel_divider(uint8_t controller, uint16_t division)
{
  if (controller > 0)
    return;

  if ((division < 1) || (division > 256))
    return;

  // Enable change.

  _regs[Lcd_divider] = _regs[Lcd_divider] | Lcd_change_enable;

  // Set the divider.

  set_field(Lcd_divider, 0xff, Lcd_divider_value, division - 1);

  // Restart clock and disable change.

  while (_regs[Lcd_divider] & Lcd_change_busy);
  _regs[Lcd_divider] = _regs[Lcd_divider] & ~Lcd_change_enable;
}



// Clock sources.

uint8_t
Cpm_x1600_chip::get_memory_source()
{
  return get_field(Ddr_divider, 0x3, Clock_source_ddr);
}

uint32_t
Cpm_x1600_chip::get_memory_source_frequency()
{
  switch (get_memory_source())
  {
    case Source_mux_main:
      return get_main_frequency();
    case Source_mux_pll_M:
      return get_pll_frequency(Pll_control_M);
    default:
      return 0;
  }
}

uint8_t
Cpm_x1600_chip::get_cpu_source()
{
  return get_field(Clock_control, 0x3, Clock_source_cpu);
}

uint32_t
Cpm_x1600_chip::get_cpu_source_frequency()
{
  switch (get_cpu_source())
  {
    case Source_mux_main:
      return get_main_frequency();
    case Source_mux_pll_M:
      return get_pll_frequency(Pll_control_M);
    default:
      return 0;
  }
}

uint8_t
Cpm_x1600_chip::get_hclock0_source()
{
  return get_field(Clock_control, 0x3, Clock_source_hclock0);
}

uint32_t
Cpm_x1600_chip::get_hclock0_source_frequency()
{
  switch (get_hclock0_source())
  {
    case Source_mux_main:
      return get_main_frequency();
    case Source_mux_pll_M:
      return get_pll_frequency(Pll_control_M);
    default:
      return 0;
  }
}

uint8_t
Cpm_x1600_chip::get_hclock2_source()
{
  return get_field(Clock_control, 0x3, Clock_source_hclock2);
}

uint32_t
Cpm_x1600_chip::get_hclock2_source_frequency()
{
  switch (get_hclock2_source())
  {
    case Source_mux_main:
      return get_main_frequency();
    case Source_mux_pll_M:
      return get_pll_frequency(Pll_control_M);
    default:
      return 0;
  }
}

void
Cpm_x1600_chip::set_hclock2_source(uint8_t source)
{
  set_field(Clock_control, 0x3, Clock_source_hclock2, source);
}

uint8_t
Cpm_x1600_chip::get_lcd_source(uint8_t controller)
{
  (void) controller;
  return get_field(Lcd_divider, 0x3, Clock_source_lcd);
}

uint32_t
Cpm_x1600_chip::get_lcd_source_frequency(uint8_t controller)
{
  switch (get_lcd_source(controller))
  {
    case Source_main:
      return get_main_frequency();
    case Source_pll_M:
      return get_pll_frequency(Pll_control_M);
    case Source_pll_E:
      return get_pll_frequency(Pll_control_E);
    default:
      return 0;
  }
}

void
Cpm_x1600_chip::set_lcd_source(uint8_t controller, uint8_t source)
{
  if (controller > 0)
    return;

  // Stop clock and enable change.

  _regs[Lcd_divider] = _regs[Lcd_divider] | Lcd_change_enable | Lcd_clock_stop;

  // Set the source.

  set_field(Lcd_divider, 0x03, Clock_source_lcd, source);

  // Restart clock and disable change.

  while (_regs[Lcd_divider] & Lcd_change_busy);
  _regs[Lcd_divider] = _regs[Lcd_divider] & ~(Lcd_change_enable | Lcd_clock_stop);
}

uint8_t
Cpm_x1600_chip::get_pclock_source()
{
  return get_hclock2_source();
}

uint32_t
Cpm_x1600_chip::get_pclock_source_frequency()
{
  return get_hclock2_source_frequency();
}

void
Cpm_x1600_chip::set_pclock_source(uint8_t source)
{
  set_hclock2_source(source);
}



// Source frequency, used by various clock sources.

uint8_t
Cpm_x1600_chip::get_main_source()
{
  return get_field(Clock_control, 0x3, Clock_source_main);
}

uint32_t
Cpm_x1600_chip::get_main_frequency()
{
  switch (get_main_source())
  {
    case Source_pll_A:
      return get_pll_frequency(Pll_control_A);
    case Source_external:
      return _exclk_freq;
    default:
      return 0;
  }
}

// Clock frequency for the CPU.

uint32_t
Cpm_x1600_chip::get_cpu_frequency()
{       
  return get_cpu_source_frequency() / get_cpu_divider();
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_x1600_chip::get_hclock0_frequency()
{
  return get_hclock0_source_frequency() / get_hclock0_divider();
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_x1600_chip::get_hclock2_frequency()
{
  return get_hclock2_source_frequency() / get_hclock2_divider();
}

// Clock frequency for slow peripherals.

uint32_t
Cpm_x1600_chip::get_pclock_frequency()
{
  return get_pclock_source_frequency() / get_pclock_divider();
}

// Clock frequency for the memory.

uint32_t
Cpm_x1600_chip::get_memory_frequency()
{
  return get_memory_source_frequency() / get_memory_divider();
}

uint32_t
Cpm_x1600_chip::get_apll_frequency()
{
  return get_pll_frequency(Pll_control_A);
}

uint32_t
Cpm_x1600_chip::get_epll_frequency()
{
  return get_pll_frequency(Pll_control_E);
}

uint32_t
Cpm_x1600_chip::get_mpll_frequency()
{
  return get_pll_frequency(Pll_control_M);
}



uint32_t
Cpm_x1600_chip::get_frequency(enum Clock_identifiers clock)
{
  switch (clock)
  {
    // NOTE: Returning only the frequency for controller 0.

    case Clock_lcd_pixel:
      return get_lcd_source_frequency(0) / get_lcd_pixel_divider(0);

    // NOTE: Consider a better error result.

    default:
      return 0;
  }
}

void
Cpm_x1600_chip::set_frequency(enum Clock_identifiers clock, uint32_t frequency)
{
  switch (clock)
  {
    // The pixel frequency is based on the selected clock source (SCLK_A, MPLL or
    // EPLL).

    case Clock_lcd_pixel:

      // Switch to the MPLL and attempt to set the divider.

      set_lcd_source(0, Source_pll_M);
      pll_enable(Pll_control_M);
      set_lcd_pixel_divider(0, get_lcd_source_frequency() / frequency);
      break;

    default:
      break;
  }
}



// C language interface functions.

void
*x1600_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 24MHz EXCLK frequency. */

  return (void *) new Cpm_x1600_chip(cpm_base, 24000000);
}

int
x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->have_clock(clock);
}

void
x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->start_clock(clock);
}

void
x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->stop_clock(clock);
}



uint8_t
x1600_cpm_get_cpu_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_cpu_divider();
}

uint8_t
x1600_cpm_get_hclock0_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock0_divider();
}

uint8_t
x1600_cpm_get_hclock2_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock2_divider();
}

uint8_t
x1600_cpm_get_lcd_pixel_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_lcd_pixel_divider();
}

uint8_t
x1600_cpm_get_memory_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_memory_divider();
}

uint8_t
x1600_cpm_get_pclock_divider(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_pclock_divider();
}



uint8_t
x1600_cpm_get_hclock0_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock0_source();
}

uint8_t
x1600_cpm_get_hclock2_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock2_source();
}

uint8_t
x1600_cpm_get_lcd_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_lcd_source();
}

uint8_t
x1600_cpm_get_memory_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_memory_source();
}

uint8_t
x1600_cpm_get_pclock_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_pclock_source();
}

void
x1600_cpm_set_pclock_source(void *cpm, uint8_t source)
{       
  static_cast<Cpm_x1600_chip *>(cpm)->set_pclock_source(source);
}



uint32_t
x1600_cpm_get_hclock0_source_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock0_source_frequency();
}

uint32_t
x1600_cpm_get_hclock2_source_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock2_source_frequency();
}

uint32_t
x1600_cpm_get_lcd_source_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_lcd_source_frequency();
}

uint32_t
x1600_cpm_get_memory_source_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_memory_source_frequency();
}

uint32_t
x1600_cpm_get_pclock_source_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_pclock_source_frequency();
}



uint8_t
x1600_cpm_get_main_source(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_main_source();
}

uint32_t
x1600_cpm_get_main_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_main_frequency();
}

uint32_t
x1600_cpm_get_cpu_frequency(void *cpm)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_cpu_frequency();
}

uint32_t
x1600_cpm_get_hclock0_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock0_frequency();
}

uint32_t
x1600_cpm_get_hclock2_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_hclock2_frequency();
}

uint32_t
x1600_cpm_get_memory_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_memory_frequency();
}

uint32_t
x1600_cpm_get_pclock_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_pclock_frequency();
}

uint32_t
x1600_cpm_get_apll_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_apll_frequency();
}

uint32_t
x1600_cpm_get_epll_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_epll_frequency();
}

uint32_t
x1600_cpm_get_mpll_frequency(void *cpm)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_mpll_frequency();
}



uint32_t
x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_frequency(clock);
}

void
x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_frequency(clock, frequency);
}

void
x1600_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_pll_parameters(Pll_control_M, multiplier, in_divider, out_divider);
}
